import pandas as pd
import numpy as np
from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.exceptions.AllCountsFilteredException import AllCountsFilteredException
from cellphonedb.src.core.exceptions.NoInteractionsFound import NoInteractionsFound
from cellphonedb.src.core.methods import cpdb_analysis_helper, cpdb_statistical_analysis_complex_method
from cellphonedb.src.core.methods import cpdb_statistical_analysis_helper


def call(meta: pd.DataFrame,
         counts: pd.DataFrame,
         counts_data: str,
         interactions: pd.DataFrame,
         genes: pd.DataFrame,
         complexes: pd.DataFrame,
         complex_compositions: pd.DataFrame,
         separator: str,
         threshold: float = 0.1,
         result_precision: int = 3
         ) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):
    core_logger.info(
        '[Non Statistical Method] Threshold:{} Precision:{}'.format(threshold,
                                                                    result_precision))

    cells_names = sorted(counts.columns)

    interactions.set_index('id_interaction', drop=True, inplace=True)
    interactions_reduced = interactions[['multidata_1_id', 'multidata_2_id']].drop_duplicates()

    complex_compositions.set_index('id_complex_composition', inplace=True, drop=True)
    # Add id multidata to counts input
    counts: pd.DataFrame = counts.merge(genes[['id_multidata', 'ensembl', 'gene_name', 'hgnc_symbol']],
                                        left_index=True, right_on=counts_data)
    counts_relations = counts[['id_multidata', 'ensembl', 'gene_name', 'hgnc_symbol']].copy()

    counts.set_index('id_multidata', inplace=True, drop=True)
    counts = counts[cells_names]
    if np.any(counts.dtypes.values != np.dtype('float32')):
        counts = counts.astype(np.float32)
    counts = counts.groupby(counts.index).mean()

    if counts.empty:
        raise AllCountsFilteredException(hint='Are you using human data?')
    # End add id multidata

    interactions_filtered, counts_filtered, complex_composition_filtered = \
        cpdb_statistical_analysis_helper.prefilters(interactions_reduced,
                                                    counts,
                                                    complexes,
                                                    complex_compositions)
    if interactions_filtered.empty:
        raise NoInteractionsFound()

    meta = meta.loc[counts.columns]

    clusters = cpdb_statistical_analysis_helper.build_clusters(meta, counts_filtered, complex_composition_filtered, skip_percent=False)
    core_logger.info('Running Real Analysis')

    cluster_interactions = cpdb_statistical_analysis_helper.get_cluster_combinations(clusters['names'])

    base_result = cpdb_statistical_analysis_helper.build_result_matrix(interactions_filtered,
                                                                       cluster_interactions,
                                                                       separator)

    mean_analysis = cpdb_statistical_analysis_helper.mean_analysis(interactions_filtered,
                                                                   clusters,
                                                                   cluster_interactions,
                                                                   base_result,
                                                                   separator)

    percent_analysis = cpdb_analysis_helper.percent_analysis(clusters,
                                                             threshold,
                                                             interactions_filtered,
                                                             cluster_interactions,
                                                             base_result,
                                                             separator)

    means_result, significant_means, deconvoluted_result = build_results(
        interactions_filtered,
        interactions,
        counts_relations,
        mean_analysis,
        percent_analysis,
        clusters['means'],
        complex_composition_filtered,
        counts,
        genes,
        result_precision,
        counts_data
    )
    return means_result, significant_means, deconvoluted_result


def build_results(interactions: pd.DataFrame,
                  interactions_original: pd.DataFrame,
                  counts_relations: pd.DataFrame,
                  mean_analysis: pd.DataFrame,
                  percent_analysis: pd.DataFrame,
                  clusters_means: pd.DataFrame,
                  complex_compositions: pd.DataFrame,
                  counts: pd.DataFrame,
                  genes: pd.DataFrame,
                  result_precision: int,
                  counts_data: str) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):
    """
    Sets the results data structure from method generated data. Results documents are defined by specs.
    """
    core_logger.info('Building results')
    interactions: pd.DataFrame = interactions_original.loc[interactions.index]
    interactions['interaction_index'] = interactions.index
    interactions = interactions.merge(counts_relations, how='left', left_on='multidata_1_id', right_on='id_multidata', )
    interactions = interactions.merge(counts_relations, how='left', left_on='multidata_2_id', right_on='id_multidata',
                                      suffixes=('_1', '_2'))
    interactions.set_index('interaction_index', inplace=True, drop=True)

    interacting_pair = cpdb_statistical_analysis_helper.interacting_pair_build(interactions)

    def simple_complex_indicator(interaction: pd.Series, suffix: str) -> str:
        """
        Add simple/complex prefixes to interaction components
        """
        if interaction['is_complex{}'.format(suffix)]:
            return 'complex:{}'.format(interaction['name{}'.format(suffix)])

        return 'simple:{}'.format(interaction['name{}'.format(suffix)])

    interactions['partner_a'] = interactions.apply(lambda interaction: simple_complex_indicator(interaction, '_1'),
                                                   axis=1)
    interactions['partner_b'] = interactions.apply(lambda interaction: simple_complex_indicator(interaction, '_2'),
                                                   axis=1)

    significant_mean_rank, significant_means = cpdb_analysis_helper.build_significant_means(
        mean_analysis,
        percent_analysis)
    significant_means = significant_means.round(result_precision)

    gene_columns = ['{}_{}'.format(counts_data, suffix) for suffix in ('1', '2')]
    gene_renames = {column: 'gene_{}'.format(suffix) for column, suffix in zip(gene_columns, ['a', 'b'])}

    # Remove useless columns
    interactions_data_result = pd.DataFrame(
        interactions[['id_cp_interaction', 'partner_a', 'partner_b', 'receptor_1', 'receptor_2', *gene_columns,
                      'annotation_strategy']].copy())

    interactions_data_result = pd.concat([interacting_pair, interactions_data_result], axis=1, sort=False)

    interactions_data_result['secreted'] = (interactions['secreted_1'] | interactions['secreted_2'])
    interactions_data_result['is_integrin'] = (interactions['integrin_1'] | interactions['integrin_2'])

    interactions_data_result.rename(
        columns={**gene_renames, 'receptor_1': 'receptor_a', 'receptor_2': 'receptor_b'},
        inplace=True)

    # Dedupe rows and filter only desired columns
    interactions_data_result.drop_duplicates(inplace=True)

    means_columns = ['id_cp_interaction', 'interacting_pair', 'partner_a', 'partner_b', 'gene_a', 'gene_b', 'secreted',
                     'receptor_a', 'receptor_b', 'annotation_strategy', 'is_integrin']

    interactions_data_result = interactions_data_result[means_columns]

    mean_analysis = mean_analysis.round(result_precision)

    # Round result decimals
    for key, cluster_means in clusters_means.items():
        clusters_means[key] = cluster_means.round(result_precision)

    # Document 2
    means_result = pd.concat([interactions_data_result, mean_analysis], axis=1, join='inner', sort=False)

    # Document 3
    significant_means_result = pd.concat([interactions_data_result, significant_mean_rank, significant_means], axis=1,
                                         join='inner', sort=False)

    # Document 5
    deconvoluted_result = cpdb_statistical_analysis_complex_method.deconvoluted_complex_result_build(
        clusters_means,
        interactions,
        complex_compositions,
        counts,
        genes,
        counts_data)

    return means_result, significant_means_result, deconvoluted_result


def deconvoluted_complex_result_build(clusters_means: dict, interactions: pd.DataFrame,
                                      complex_compositions: pd.DataFrame, counts: pd.DataFrame,
                                      genes: pd.DataFrame, counts_data: str) -> pd.DataFrame:
    genes_counts = list(counts.index)
    genes_filtered = genes[genes[counts_data].apply(lambda gene: gene in genes_counts)]

    deconvoluted_complex_result_1 = deconvolute_complex_interaction_component(complex_compositions, genes_filtered,
                                                                              interactions, '_1', counts_data)
    deconvoluted_simple_result_1 = deconvolute_interaction_component(interactions, '_1', counts_data)

    deconvoluted_complex_result_2 = deconvolute_complex_interaction_component(complex_compositions, genes_filtered,
                                                                              interactions, '_2', counts_data)
    deconvoluted_simple_result_2 = deconvolute_interaction_component(interactions, '_2', counts_data)

    deconvoluted_result = deconvoluted_complex_result_1.append(
        [deconvoluted_simple_result_1, deconvoluted_complex_result_2, deconvoluted_simple_result_2], sort=False)

    deconvoluted_result.set_index('gene', inplace=True)

    cluster_counts = pd.DataFrame(index=deconvoluted_result.index)

    for key, cluster_means in clusters_means.items():
        cluster_counts[key] = cluster_means

    cluster_counts = cluster_counts.reindex(sorted(cluster_counts.columns), axis=1)

    # Here we sort and filter unwanted columns
    deconvoluted_columns = ['gene_name', 'name', 'is_complex', 'protein_name', 'complex_name', 'id_cp_interaction']

    deconvoluted_result = deconvoluted_result[deconvoluted_columns]
    deconvoluted_result.rename({'name': 'uniprot'}, axis=1, inplace=True)

    deconvoluted_result = pd.concat([deconvoluted_result, cluster_counts], axis=1, join='inner', sort=False)

    deconvoluted_result.reset_index(inplace=True)
    deconvoluted_result.drop(columns='gene', inplace=True)
    return deconvoluted_result


def deconvolute_interaction_component(interactions, suffix, counts_data):
    interactions = interactions[~interactions['is_complex{}'.format(suffix)]]
    deconvoluted_result = pd.DataFrame()
    deconvoluted_result['gene'] = interactions['{}{}'.format(counts_data, suffix)]

    deconvoluted_result[['protein_name', 'gene_name', 'name', 'is_complex', 'id_cp_interaction', 'receptor']] = \
        interactions[['protein_name{}'.format(suffix), 'gene_name{}'.format(suffix), 'name{}'.format(suffix),
                      'is_complex{}'.format(suffix), 'id_cp_interaction', 'receptor{}'.format(suffix)]]
    deconvoluted_result['complex_name'] = np.nan

    return deconvoluted_result


def deconvolute_complex_interaction_component(complex_compositions, genes_filtered, interactions, suffix, counts_data):
    deconvoluted_result = pd.DataFrame()
    component = pd.DataFrame()
    component[counts_data] = interactions['{}{}'.format(counts_data, suffix)]
    component[['protein_name', 'gene_name', 'name', 'is_complex', 'id_cp_interaction', 'id_multidata', 'receptor']] = \
        interactions[
            ['protein_name{}'.format(suffix), 'gene_name{}'.format(suffix),
             'name{}'.format(suffix), 'is_complex{}'.format(suffix), 'id_cp_interaction',
             'id_multidata{}'.format(suffix), 'receptor{}'.format(suffix)]]

    deconvolution_complex = pd.merge(complex_compositions, component, left_on='complex_multidata_id',
                                     right_on='id_multidata')
    deconvolution_complex = pd.merge(deconvolution_complex, genes_filtered, left_on='protein_multidata_id',
                                     right_on='protein_multidata_id', suffixes=['_complex', '_simple'])

    deconvoluted_result['gene'] = deconvolution_complex['{}_simple'.format(counts_data)]

    deconvoluted_result[
        ['protein_name', 'gene_name', 'name', 'is_complex', 'id_cp_interaction', 'receptor', 'complex_name']] = \
        deconvolution_complex[
            ['protein_name_simple', 'gene_name_simple', 'name_simple',
             'is_complex_complex', 'id_cp_interaction', 'receptor_simple', 'name_complex']]

    return deconvoluted_result



import flask
from cellphonedb.src.api_endpoints.web_api.web_api_endpoint_base import WebApiEndpointBase
from cellphonedb.src.database.manager import DatabaseVersionManager


class WebEndpointQueryDatabase(WebApiEndpointBase):
    def get(self):
        response = DatabaseVersionManager.list_local_versions()

        return flask.jsonify(response)



import pandas as pd

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.core.generators import generator_helper


def protein_generator(uniprot_db: pd.DataFrame,
                      curated_proteins: pd.DataFrame,
                      user_protein: pd.DataFrame,
                      default_values: dict,
                      default_types: dict,
                      result_columns: list,
                      log_path: str) -> pd.DataFrame:
    merged_proteins = _merge_proteins(uniprot_db, curated_proteins, default_values, default_types, result_columns,
                                      log_path, quiet=True)

    # TODO: Add missing mandatory check
    if isinstance(user_protein, pd.DataFrame) and not user_protein.empty:
        merged_proteins = _merge_proteins(merged_proteins, user_protein, default_values, default_types, result_columns,
                                          log_path, quiet=False)

    return merged_proteins


def _merge_proteins(base_protein: pd.DataFrame,
                    additional: pd.DataFrame,
                    default_values: dict,
                    default_types: dict,
                    result_columns: list,
                    log_file: str,
                    quiet: bool = False) -> pd.DataFrame:
    additional = additional.copy()

    # Here we set defaults for uniprot & curated data
    base_protein = generator_helper.set_defaults(base_protein, default_values, quiet)
    additional = generator_helper.set_defaults(additional, default_values, quiet)

    # we will only use these columns
    additional = additional[result_columns]
    base_protein = base_protein[result_columns]

    # Type casting to ensure they are equal
    additional = additional.astype(default_types)
    base_protein = base_protein.astype(default_types)

    join_key = 'uniprot'

    merged_protein = base_protein.append(additional, ignore_index=True, sort=False).drop_duplicates()

    if not quiet and merged_protein.duplicated(join_key).any():
        app_logger.warning('There are differences between merged files: logged to {}'.format(log_file))

        log = merged_protein[merged_protein.duplicated(join_key, keep=False)].sort_values(join_key)
        log.to_csv(log_file, index=False, sep='\t')

    merged_protein.drop_duplicates(join_key, keep='last', inplace=True)
    return merged_protein



import pandas as pd

from cellphonedb.src.core.models.multidata import multidata_properties


def is_cellphonedb_interactor(interaction: pd.Series, suffixes=('_1', '_2')) -> bool:
    if interaction['annotation_strategy'] == 'curated':
        return True

    if interaction['annotation_strategy'] == 'user_curated':
        return True

    if interaction['id_multidata{}'.format(suffixes[0])] == interaction['id_multidata{}'.format(suffixes[1])]:
        return False

    if interaction['annotation_strategy'] == 'guidetopharmacology.org':
        return True

    if multidata_properties.can_be_receptor(interaction, suffixes[0]) and \
            multidata_properties.can_be_ligand(interaction, suffixes[1]):
        return True

    if multidata_properties.can_be_receptor(interaction, suffixes[1]) and \
            multidata_properties.can_be_ligand(interaction, suffixes[0]):
        return True

    return False



import re

import pandas as pd


def autocomplete_query(genes: pd.DataFrame, multidatas: pd.DataFrame, partial_element: pd.DataFrame) -> pd.DataFrame:
    values = _partial_filter(genes, 'ensembl', partial_element)

    by_protein_name = _partial_filter(genes, 'protein_name', partial_element)
    by_gene_name = _partial_filter(genes, 'gene_name', partial_element)

    with_hgnc_symbol = genes.dropna(subset=['hgnc_symbol'])
    by_hgnc_symbol = _partial_filter(with_hgnc_symbol, 'hgnc_symbol', partial_element)

    by_name = _partial_filter(multidatas, 'name', partial_element)

    values = values.append(by_protein_name, ignore_index=True)
    values = values.append(by_gene_name, ignore_index=True)
    values = values.append(by_hgnc_symbol, ignore_index=True)
    values = values.append(by_name, ignore_index=True)

    result = pd.DataFrame(data=values, columns=['value']).drop_duplicates()

    return result


def _partial_filter(input_data, name, partial_element):
    matching = input_data[input_data[name].str.contains(partial_element, flags=re.IGNORECASE)][name]
    return matching



import pandas as pd


def get_involved_complex_composition_from_protein(proteins: pd.DataFrame,
                                                  complex_compositions: pd.DataFrame) -> pd.DataFrame:
    """
    Returns a table of complex with all proteins declared in proteins.
    """
    multidata_protein = list(proteins.index)

    complex_counts_filtered = complex_compositions[
        complex_compositions['protein_multidata_id'].apply(lambda protein_multidata: protein_multidata in multidata_protein)]

    if complex_counts_filtered.empty:
        return pd.DataFrame(columns=complex_compositions.columns)

    def all_protein_involved(complex: pd.Series) -> bool:
        number_proteins_in_counts = len(
            complex_counts_filtered[
                complex_counts_filtered['complex_multidata_id'] == complex['complex_multidata_id']])

        if number_proteins_in_counts < complex['total_protein']:
            return False

        return True

    complex_composition_complete = complex_counts_filtered[
        complex_counts_filtered.apply(all_protein_involved, axis=1)]

    return complex_composition_complete



import pandas as pd


def merge_genes_from_uniprot_ensembl_db(ensembls: pd.DataFrame, proteins: pd.DataFrame,
                                        uniprots: pd.DataFrame) -> pd.DataFrame:
    uniprots_filtered = merge_genes_cellphone(proteins, uniprots)
    merged_genes = pd.merge(uniprots_filtered, ensembls, left_on=0, right_on='Gene name')

    result = _merge_ensembl_uniprots(merged_genes)
    result = result[~result['gene_name'].str.contains('HLA')]
    result = result[result['uniprot'].apply(lambda uniprot: uniprot in proteins['uniprot'].tolist())]
    result = result.sort_values(by=list(result.columns.values))

    return result


def _merge_ensembl_uniprots(merged_genes: pd.DataFrame) -> pd.DataFrame:
    merged_from_uniprot = pd.DataFrame()

    merged_from_uniprot[['ensembl', 'uniprot', 'gene_name', 'hgnc_symbol']] = merged_genes[
        ['Gene stable ID', 'Entry', 'Gene name', 'HGNC symbol']]

    merged_from_ensembl = pd.DataFrame()
    merged_from_ensembl[['ensembl', 'uniprot', 'gene_name', 'hgnc_symbol']] = merged_genes[[
        'Gene stable ID', 'UniProtKB/Swiss-Prot ID', 'Gene name', 'HGNC symbol']]

    result = merged_from_uniprot.append(merged_from_ensembl, sort=False).reset_index(drop=True)

    result.dropna(subset=['ensembl', 'uniprot', 'gene_name'], inplace=True)

    check_empty_hgnc(result)

    result['hgnc_symbol'] = result.apply(
        lambda gene: gene['hgnc_symbol'] if gene['hgnc_symbol'] == '' else gene['gene_name'], axis=1)

    result = result.drop_duplicates().reset_index(drop=True)
    return result


def check_empty_hgnc(genes: pd.DataFrame) -> None:
    empty_hgnc = genes[pd.isnull(genes['hgnc_symbol'])].drop_duplicates()
    if not empty_hgnc.empty:
        print('SOME GENES DO NOT HAVE HGNC')
        print(empty_hgnc.to_csv(index=False))


def merge_genes_cellphone(proteins: pd.DataFrame, uniprots: pd.DataFrame) -> pd.DataFrame:
    uniprots_merged = pd.merge(proteins, uniprots, left_on='uniprot', right_on='Entry', indicator=True, how='outer')

    only_in_protein = uniprots_merged[uniprots_merged['_merge'] == 'left_only']

    if not only_in_protein.empty:
        print('SOME PROTEINS DIDNT EXIST IN UNIPROT GENE DATABASE')
        print(only_in_protein['uniprot'].tolist())

    uniprots_filtered = uniprots_merged[uniprots_merged['_merge'] == 'both']
    uniprots_filtered = _deconvolute_genenames(uniprots_filtered)

    return uniprots_filtered


def _deconvolute_genenames(uniprots_filtered: pd.DataFrame) -> pd.DataFrame:
    '''
    Deconvolutes gene names in uniprots database
    ie:
    input:
    Entry   Gene names
    A       a b c
    B       d

    result:

    Entry   Gene name
    A       a
    A       b
    A       c
    B       d
    '''
    uniprots_columns = ['Entry', 'Gene names', 'Ensembl transcript']
    uniprots_filtered = uniprots_filtered[uniprots_columns]

    deconvoluted = uniprots_filtered['Gene names'].str.split(' ').apply(pd.Series)
    deconvoluted.index = uniprots_filtered.set_index(uniprots_columns).index
    deconvoluted = deconvoluted.stack().reset_index(uniprots_columns).reset_index(drop=True)

    return deconvoluted
#



import pandas as pd

from cellphonedb.tools.generate_data.mergers import merge_interactions
from cellphonedb.tools.tools_helper import normalize_interactions


def add_curated(interactions: pd.DataFrame, interaction_curated: pd.DataFrame) -> pd.DataFrame:
    interactions.rename(index=str, columns={'uniprot_1': 'partner_a', 'uniprot_2': 'partner_b'},
                        inplace=True)

    interactions_curated_normalized = normalize_interactions(interaction_curated, 'partner_a',
                                                             'partner_b')
    curated_duplicated = interactions_curated_normalized[
        interactions_curated_normalized.duplicated(['partner_a', 'partner_b'], keep=False)]

    if not curated_duplicated.empty:
        print(
            'WARNING - Some curated interactions are duplicated.')
        print(interaction_curated.iloc[
            curated_duplicated.sort_values(['partner_a', 'partner_b']).index.values].to_csv(
            index=False))
    interactions.rename(index=str, columns={'protein_1': 'partner_a', 'protein_2': 'partner_b'},
                        inplace=True)

    interaction_curated_result = merge_interactions.merge_iuphar_other_and_curated_interactions(interactions,
                                                                                                interaction_curated)

    return interaction_curated_result



from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.complex.web_endpoint_query_complex_deconvoluted import \
    WebEndpointQueryComplexDeconvoluted


def add(api, prefix=''):
    api.add_resource(WebEndpointQueryComplexDeconvoluted, '{}/deconvoluted'.format(prefix))



import pandas as pd

from cellphonedb.src.core.methods import cpdb_statistical_analysis_complex_method


def call(meta: pd.DataFrame,
         count: pd.DataFrame,
         counts_data: str,
         interactions: pd.DataFrame,
         genes: pd.DataFrame,
         complex_expanded: pd.DataFrame,
         complex_composition: pd.DataFrame,
         iterations: int,
         threshold: float,
         threads: int,
         debug_seed: int,
         result_precision: int,
         pvalue: float,
         separator: str
         ) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame):

    pvalues, means, significant_means, deconvoluted = \
        cpdb_statistical_analysis_complex_method.call(meta.copy(),
                                                      count.copy(),
                                                      counts_data,
                                                      interactions,
                                                      genes,
                                                      complex_expanded,
                                                      complex_composition,
                                                      pvalue,
                                                      separator,
                                                      iterations,
                                                      threshold,
                                                      threads,
                                                      debug_seed,
                                                      result_precision,
                                                      )


    max_rank = significant_means['rank'].max()
    significant_means['rank'] = significant_means['rank'].apply(lambda rank: rank if rank != 0 else (1 + max_rank))
    significant_means.sort_values('rank', inplace=True)

    return deconvoluted, means, pvalues, significant_means



from unittest import TestCase

import pandas as pd

from cellphonedb.src.core.models.multidata import multidata_properties


class TestPropertiesMultidata(TestCase):
    def test_can_be_ligand(self):
        multidatas = pd.DataFrame(multidatas_fixtures)

        for index, multidata in multidatas.iterrows():
            self.assertEqual(multidata_properties.can_be_ligand(multidata), multidata['can_be_ligand'],
                             'Multidata {} cant be ligand didnt match'.format(multidata['id']))

    def test_can_be_receptor(self):
        multidatas = pd.DataFrame(multidatas_fixtures)
        for index, multidata in multidatas.iterrows():
            self.assertEqual(multidata_properties.can_be_receptor(multidata), multidata['can_be_receptor'],
                             'Multidata {} can be receptor didnt match')


multidatas_fixtures = [
    {
        'id': 1,
        'receptor': True,
        'secreted': True,
        'other': False,
        'transmembrane': False,
        'extracellular': False,
        'is_receptor': False,
        'secreted_highlight': True,
        'can_be_ligand': True,
        'can_be_receptor': True

    },
    {
        'id': 2,
        'receptor': True,
        'secreted': False,
        'other': True,
        'transmembrane': True,
        'extracellular': True,
        'is_receptor': True,
        'secreted_highlight': False,
        'can_be_ligand': False,
        'can_be_receptor': False

    },
    {
        'id': 3,
        'receptor': True,
        'secreted': False,
        'other': False,
        'transmembrane': False,
        'extracellular': False,
        'is_receptor': False,
        'secreted_highlight': False,
        'can_be_ligand': False,
        'can_be_receptor': True
    },
    {
        'id': 4,
        'receptor': False,
        'secreted': False,
        'other': False,
        'transmembrane': True,
        'extracellular': False,
        'is_receptor': False,
        'secreted_highlight': False,
        'can_be_ligand': False,
        'can_be_receptor': False
    },
]



from sqlalchemy import Column, Integer, String, ForeignKey

from cellphonedb.src.core.database.sqlalchemy_models import Base


class Gene(Base):
    __tablename__ = 'gene_table'
    id_gene = Column(Integer, nullable=False, primary_key=True)

    ensembl = Column(String, nullable=False)
    gene_name = Column(String, nullable=False)
    hgnc_symbol = Column(String)

    protein_id = Column(Integer, ForeignKey('protein_table.id_protein'), nullable=False)



import pandas as pd


def call(multidatas: pd.DataFrame, interactions: pd.DataFrame):
    multidatas_interactors_ids = interactions['multidata_1_id'].append(
        interactions['multidata_2_id']).drop_duplicates().reset_index(drop=True).tolist()

    multidatas_filtered = multidatas[multidatas['id_multidata'].apply(lambda id: id in multidatas_interactors_ids)]

    return multidatas_filtered[
        ['name', 'protein_name', 'transmembrane', 'secreted', 'peripheral', 'integrin',
         'secreted_highlight',
         'secreted_desc', 'other', 'other_desc', 'tags', 'tags_description', 'tags_reason']]



import os

from sqlalchemy import create_engine

from cellphonedb.src.core.Cellphonedb import Cellphonedb
from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database.Database import Database
from cellphonedb.src.core.database.DatabaseManager import DatabaseManager
from cellphonedb.src.core.database.sqlalchemy_models import Base
from cellphonedb.src.core.database.sqlalchemy_repository.ComplexRepository import ComplexRepository
from cellphonedb.src.core.database.sqlalchemy_repository.GeneRepository import GeneRepository
from cellphonedb.src.core.database.sqlalchemy_repository.InteractionRepository import InteractionRepository
from cellphonedb.src.core.database.sqlalchemy_repository.MultidataRepository import MultidataRepository
from cellphonedb.src.core.database.sqlalchemy_repository.ProteinRepository import ProteinRepository


class CellphonedbSqlalchemy(Cellphonedb):
    def __init__(self, config: dict, database_file=None, collecting=False):
        core_logger.setLevel(config['logger']['level'])
        core_logger.info('Initializing SqlAlchemy CellPhoneDB Core')

        if database_file:
            if not collecting and not os.path.exists(database_file):
                raise Exception('Given database file {} does not exist'.format(database_file))

            uri = self._build_sqlite_uri(database_file)
        else:
            # todo: Improve config stuff
            uri = self._build_uri(config)

        core_logger.debug('Database Uri: {}'.format(uri))

        engine = create_engine(uri)
        database = Database(engine)
        database.base_model = Base
        database_manager = DatabaseManager(None, database)
        # TODO: Auto-load repositories
        database_manager.add_repository(ComplexRepository)
        database_manager.add_repository(GeneRepository)
        database_manager.add_repository(InteractionRepository)
        database_manager.add_repository(MultidataRepository)
        database_manager.add_repository(ProteinRepository)
        Cellphonedb.__init__(self, database_manager, config)

    @staticmethod
    def _build_sqlite_uri(database_file):
        path = os.path.realpath(os.path.expanduser(database_file))
        core_logger.info('Using custom database at {}'.format(path))

        return 'sqlite:///{}'.format(path)

    @staticmethod
    def _build_uri(config):
        if config['sqlalchemy']['db_core']:

            file_path = os.path.dirname(os.path.realpath(__file__))

            if not config['sqlalchemy']['uri']:
                return 'sqlite:///{}/cellphone.db'.format(file_path)

            return 'sqlite:///{}/{}'.format(file_path, config['sqlalchemy']['uri'])
        else:
            return config['sqlalchemy']['uri']



import hashlib

import pandas as pd


def interaction(interaction: pd.Series, suffixes=('_x', '_y')) -> str:
    prefix = 'CPI'

    simple_complex_indicator_1 = _complex_indicator(interaction['is_complex{}'.format(suffixes[0])])
    simple_complex_indicator_2 = _complex_indicator(interaction['is_complex{}'.format(suffixes[1])])

    hash = _calculate_interaction_hash(interaction['name{}'.format(suffixes[0])],
                                       interaction['name{}'.format(suffixes[1])])

    control_digit = '0'

    return '{}-{}{}{}{}'.format(prefix, simple_complex_indicator_1, simple_complex_indicator_2, control_digit,
                                hash.upper())


def _complex_indicator(is_complex: bool) -> str:
    if is_complex:
        return 'C'
    else:
        return 'S'


def _calculate_interaction_hash(element_1: str, element_2: str, digits: int = 8) -> str:
    return hashlib.sha1((element_1 + element_2).encode('utf-8')).hexdigest()[:digits]



from setuptools import setup, find_packages

with open("README.md", "rt", encoding="utf-8") as fh:
    long_description_readme_md = fh.read()

with open("requirements.txt", "rt", encoding="utf-8") as fh:
    install_requirements_txt = [line.strip() for line in fh.readlines()]

setup(
    name='CellPhoneDB',
    author='TeichLab',
    author_email='contact@cellphonedb.org',
    version='2.1.8b3',
    description='Inferring cell-cell communication',
    long_description=long_description_readme_md,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    maintainer='TeichLab',
    maintainer_email='contact@cellphonedb.org',
    url='https://cellphonedb.org',
    license='MIT',
    exclude_package_data={'': ['tools']},
    entry_points={
        'console_scripts':
            [
                'cellphonedb = cellphonedb.cellphonedb_cli:cli'
            ]
    },
    install_requires=install_requirements_txt,
)



import json
import os
from datetime import datetime
from io import StringIO

import pandas as pd
import requests

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import data_dir
from cellphonedb.tools.generate_data.getters.get_imex import CouldNotFetchFromApiException
from cellphonedb.tools.tools_helper import add_to_meta


def call(downloads_path: str, fetch: bool, save_backup: bool = True) -> pd.DataFrame:
    url = 'http://www.guidetopharmacology.org/DATA/interactions.csv'

    compression = 'xz'
    file_name = 'iuphar_interaction_raw.csv.{}'.format(compression)
    download_file_path = os.path.join(downloads_path, file_name)

    def best_path():
        saved_file_path = os.path.join(data_dir, 'sources', file_name)

        if os.path.exists(download_file_path):
            return download_file_path
        if os.path.exists(saved_file_path):
            return saved_file_path

        app_logger.error('Could not find local source for iuphar')
        exit(1)

    try:
        if fetch:
            response = requests.get(url)

            if response.text:
                s = StringIO(response.text)
                df = pd.read_csv(s, dtype=str)

                df.drop_duplicates(inplace=True)
                if save_backup:
                    df.to_csv(download_file_path, index=False, compression=compression)
                    add_to_meta(file_name, os.path.join(downloads_path, 'meta.json'))

                return df
            else:
                if response.status_code != 200:
                    raise CouldNotFetchFromApiException()
        else:
            app_logger.warning('Using local version for iuphar')
            df = pd.read_csv(best_path(), compression=compression, dtype=str)
            return df
    except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
        app_logger.warning('Could not fetch remote source for iuphar, using local backup')
        df = pd.read_csv(best_path(), compression=compression)
        return df



from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart

from flask_restful import Resource

import json


class WebApiEndpointBase(Resource):
    def __init__(self):
        self._msg = MIMEMultipart('form-data')
        self._errors = []
        self._attachments = []

    def attach_error(self, error: json):
        self._errors.append(error)

    def _attach_csv(self, file_to_send, filename):
        attachment = MIMEBase('text', 'csv')
        attachment.set_payload(file_to_send)

        attachment.add_header("Content-Disposition", "attachment", filename=filename)

        self._attachments.append(attachment)

    def _attach_table(self, file_to_send, filename):
        attachment = MIMEBase('text', 'tab-separated-values')
        attachment.set_payload(file_to_send)

        attachment.add_header("Content-Disposition", "attachment", filename=filename)

        self._attachments.append(attachment)

    def _attach_json(self, data, at_first=False):
        attachment = MIMEBase('application', 'json')
        attachment.set_payload(json.dumps(data))

        if at_first:
            self._attachments = [attachment] + self._attachments

        else:
            self._attachments.append(attachment)

    def _attach_status(self, data={}):
        status = {"data": {}}

        if self._errors:
            status['errors'] = self._errors

        self._attach_json(status, at_first=True)

    def _commit_attachments(self):
        self._attach_status()

        for attach in self._attachments:
            self._msg.attach(attach)



from flask import request, Response

from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.api_endpoints.web_api.web_api_endpoint_base import WebApiEndpointBase


class WebEndpointQueryInteractionGene(WebApiEndpointBase):
    @staticmethod
    def get():
        columns = request.args.get('columns')

        columns = columns.split(',') if columns else None

        genes = cellphonedb_app.cellphonedb.query.get_all_genes(columns)
        genes = genes.to_json(orient='records')

        response = Response(genes, content_type='application/json')

        return response



import pandas as pd

from cellphonedb.src.core.methods import cpdb_analysis_complex_method


def call(meta: pd.DataFrame,
         counts: pd.DataFrame,
         counts_data: str,
         interactions: pd.DataFrame,
         genes: pd.DataFrame,
         complexes: pd.DataFrame,
         complex_compositions: pd.DataFrame,
         separator: str,
         threshold: float = 0.1,
         result_precision: int = 3) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):
    means, significant_means, deconvoluted = \
        cpdb_analysis_complex_method.call(meta.copy(),
                                          counts.copy(),
                                          counts_data,
                                          interactions.copy(),
                                          genes,
                                          complexes,
                                          complex_compositions,
                                          separator,
                                          threshold,
                                          result_precision)
    max_rank = significant_means['rank'].max()
    significant_means['rank'] = significant_means['rank'].apply(lambda rank: rank if rank != 0 else (1 + max_rank))
    significant_means.sort_values('rank', inplace=True)

    return means, significant_means, deconvoluted



from unittest import TestCase

import pandas as pd

from cellphonedb.src.core.Cellphonedb import data_test_dir
from cellphonedb.src.core.models.interaction import interaction_filter
from cellphonedb.utils import dataframe_functions


class TestInteractionFilters(TestCase):
    def test_filter_by_integrin(self):
        proteins = pd.read_csv('{}/filter_interaction/filter_interaction_integrin_proteins.csv'.format(data_test_dir))
        interactions = pd.read_csv('{}/filter_interaction/filter_interaction_interactions.csv'.format(data_test_dir))
        expected_result = pd.read_csv(
            '{}/filter_interaction/filter_interaction_interactions_filtered.csv'.format(data_test_dir))

        result = interaction_filter.filter_by_receptor_ligand_integrin(proteins, interactions)

        self.assertEqual(len(result), len(expected_result[expected_result['enabled']]))
        result_matches = True
        for index, interaction in expected_result.iterrows():
            if interaction['enabled']:
                if len(result[(interaction['id_interaction'] == result['id_interaction']) &
                              (interaction['id_multidata_receptors'] == result['id_multidata_receptors']) &
                              (interaction['id_multidata_ligands'] == result['id_multidata_ligands'])
                       ]) != 1:
                    result_matches = False

        self.assertTrue(result_matches)

    def test_filter_by_multidatas(self):
        multidatas = pd.read_csv(
            '{}/filter_interaction/filter_interaction_multidatas_multidata.csv'.format(data_test_dir))
        interactions = pd.read_csv(
            '{}/filter_interaction/filter_interaction_multidatas_interaction.csv'.format(data_test_dir))

        result = interaction_filter.filter_by_multidatas(multidatas, interactions)
        expected_result = interactions[interactions['test_both_enabled']]

        self.assertTrue(dataframe_functions.dataframes_has_same_data(result, expected_result))

    def test_filter_by_any_multidata(self):

        multidatas = pd.read_csv(
            '{}/filter_interaction/filter_interaction_multidatas_multidata.csv'.format(data_test_dir))
        interactions = pd.read_csv(
            '{}/filter_interaction/filter_interaction_multidatas_interaction.csv'.format(data_test_dir))

        result = interaction_filter.filter_by_any_multidatas(multidatas, interactions)
        expected_result = interactions[interactions['test_any_enabled']]

        self.assertTrue(dataframe_functions.dataframes_has_same_data(result, expected_result))



import pandas as pd


def can_be_receptor(multidata: pd.Series, suffix: str = '') -> bool:
    if multidata['receptor{}'.format(suffix)] and \
            not multidata['other{}'.format(suffix)]:
        return True
    return False


def can_be_ligand(multidata: pd.Series, suffix: str = '') -> bool:
    if multidata['secreted_highlight{}'.format(suffix)]:
        return True
    return False



import pandas as pd

from cellphonedb.src.core.database.Repository import Repository
from cellphonedb.src.core.database.sqlalchemy_models.db_model_multidata import Multidata
from cellphonedb.src.core.database.sqlalchemy_models.db_model_protein import Protein


class ProteinRepository(Repository):
    name = 'protein'

    def get_all(self) -> pd.DataFrame:
        protein_query = self.database_manager.database.session.query(Protein)
        protein = pd.read_sql(protein_query.statement, self.database_manager.database.session.bind)

        return protein

    def get_all_expanded(self) -> pd.DataFrame:
        protein_multidata_join = Protein.protein_multidata_id == Multidata.id_multidata
        protein_query = self.database_manager.database.session.query(Protein, Multidata).join(Multidata,
                                                                                              protein_multidata_join)
        protein = pd.read_sql(protein_query.statement, self.database_manager.database.session.bind)

        return protein

    def get_all_name_id(self) -> pd.DataFrame:
        query_multidatas = self.database_manager.database.session.query(Protein.id_protein, Multidata.name).join(
            Multidata)
        multidatas = pd.read_sql(query_multidatas.statement, self.database_manager.database.session.bind)

        return multidatas

    def get_protein_multidata_by_uniprot(self, uniprot: str) -> pd.DataFrame:
        protein_query = self.database_manager.database.session.query(Protein, Multidata).join(Multidata).filter_by(
            name=uniprot).limit(1)
        protein = pd.read_sql(protein_query.statement, self.database_manager.database.session.bind)

        if not protein.empty:
            return protein.iloc[0, :]
        return None

    def add_proteins(self, proteins: pd.DataFrame, multidatas: pd.DataFrame):
        multidatas.to_sql(name='multidata_table', if_exists='append', con=self.database_manager.database.engine,
                          index=False,
                          chunksize=50)

        multidata_query = self.database_manager.database.session.query(Multidata.id_multidata, Multidata.name)
        multidatas_db = pd.read_sql(multidata_query.statement, self.database_manager.database.session.bind)
        multidatas_db.rename(index=str, columns={'id_multidata': 'protein_multidata_id'}, inplace=True)
        proteins_to_add = pd.merge(proteins, multidatas_db, on='name')
        proteins_to_add.drop('name', inplace=True, axis=1)

        proteins_to_add.to_sql(name='protein_table', if_exists='append', con=self.database_manager.database.engine,
                               index=False, chunksize=50)



import pandas as pd


def filter_by_gene(cluster_counts: pd.DataFrame, genes: pd.DataFrame, counts_data: str = 'ensembl') -> pd.DataFrame:
    """
    Returns only the clusters in genes
    """

    right_column = counts_data

    clusters_filtered = pd.merge(cluster_counts, genes, left_on='gene', right_on=right_column)

    clusters_filtered.drop(right_column, inplace=True, axis=1)

    return clusters_filtered


def filter_empty_cluster_counts(counts: pd.DataFrame, clusters_names: list) -> pd.DataFrame:
    """
    Removes counts with all values to zero
    """
    if counts.empty:
        return counts

    filtered_counts = counts[counts[clusters_names].apply(lambda row: row.sum() > 0, axis=1)]
    return filtered_counts



class NoInteractionsFound(Exception):
    def __init__(self, description: str = None, hint: str = None):
        super(NoInteractionsFound, self).__init__('No CellPhoneDB interacions found in this input.')
        self.description = description
        self.hint = hint



import pandas as pd

from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database.Repository import Repository
from cellphonedb.src.core.database.sqlalchemy_models.db_model_gene import Gene
from cellphonedb.src.core.database.sqlalchemy_models.db_model_multidata import Multidata
from cellphonedb.src.core.database.sqlalchemy_models.db_model_protein import Protein
from cellphonedb.src.core.utils import filters


class GeneRepository(Repository):
    name = 'gene'

    def get_all(self):
        query = self.database_manager.database.session.query(Gene)
        result = pd.read_sql(query.statement, self.database_manager.database.session.bind)

        return result

    def get_all_expanded(self):
        protein_multidata_join = Protein.protein_multidata_id == Multidata.id_multidata
        gene_protein_join = Gene.protein_id == Protein.id_protein
        query = self.database_manager.database.session.query(Gene, Protein, Multidata).join(
            Protein, gene_protein_join).join(Multidata, protein_multidata_join)

        result = pd.read_sql(query.statement, self.database_manager.database.session.bind)

        return result

    def add(self, genes: pd.DataFrame):
        query_multidatas = self.database_manager.database.session.query(Protein.id_protein, Multidata.name).join(
            Multidata)
        multidatas = pd.read_sql(query_multidatas.statement, self.database_manager.database.session.bind)

        genes = self._blend_multidata(genes, ['name'], multidatas)

        genes.rename(index=str, columns={'id_protein': 'protein_id'}, inplace=True)
        genes = filters.remove_not_defined_columns(genes, self.database_manager.get_column_table_names('gene_table'))

        genes.to_sql(name='gene_table', if_exists='append', con=self.database_manager.database.engine, index=False,
                     chunksize=50)

    @staticmethod
    def _blend_multidata(original_data: pd.DataFrame, original_column_names: list, multidatas: pd.DataFrame,
                         quiet: bool = False) -> pd.DataFrame:
        """
        Merges dataframe with multidata names in multidata ids
        """
        if quiet:
            core_logger.debug('Blending proteins in quiet mode')

        result = GeneRepository.blend_dataframes(original_data, original_column_names, multidatas, 'name', 'multidata')

        return result



from functools import reduce
from typing import List

import pandas as pd


def call(element: str, interactions: pd.DataFrame, complex_composition: pd.DataFrame) -> pd.DataFrame:
    elements = [name.strip() for name in element.split(',')]
    related_complexes = _find_complex_by_elements(elements, complex_composition)

    elements = elements + related_complexes

    interactions_filtered = pd.DataFrame()

    interactions_filtered = _find_interactions_by_elements(elements, interactions, interactions_filtered)

    if interactions_filtered.empty:
        return interactions_filtered

    interactions_filtered = _build_result(interactions_filtered)

    return interactions_filtered.drop_duplicates()


def _build_result(interactions: pd.DataFrame) -> pd.DataFrame:
    def simple_complex_indicator(interaction: pd.Series, suffix: str) -> str:
        if interaction['is_complex{}'.format(suffix)]:
            return 'complex:{}'.format(interaction['name{}'.format(suffix)])

        return 'simple:{}'.format(interaction['name{}'.format(suffix)])

    interactions['partner_a'] = interactions.apply(
        lambda interaction: simple_complex_indicator(interaction, '_a'), axis=1)
    interactions['partner_b'] = interactions.apply(
        lambda interaction: simple_complex_indicator(interaction, '_b'), axis=1)

    interactions = interactions[
        ['id_cp_interaction', 'partner_a', 'partner_b', 'gene_name_a', 'gene_name_b', 'ensembl_a', 'ensembl_b',
         'annotation_strategy']]
    return interactions


def _find_interactions_by_elements(elements: list, interactions: pd.DataFrame,
                                   interactions_filtered: pd.DataFrame) -> pd.DataFrame:
    for element in elements:
        element_interactions = interactions[(interactions['name_a'] == element) |
                                            (interactions['name_b'] == element) |
                                            (interactions['gene_name_a'] == element) |
                                            (interactions['gene_name_b'] == element) |
                                            (interactions['protein_name_a'] == element) |
                                            (interactions['protein_name_b'] == element) |
                                            (interactions['ensembl_a'] == element) |
                                            (interactions['ensembl_b'] == element)].copy()

        interactions_filtered = interactions_filtered.append(element_interactions)

    return interactions_filtered


def _find_complex_by_elements(elements: List[str], complex_composition: pd.DataFrame) -> List[str]:
    all_results = [_find_complex_by_element(element, complex_composition) for element in elements]

    return list(set(reduce((lambda carry, current: carry + current), all_results)))


def _find_complex_by_element(element: str, complex_composition: pd.DataFrame) -> List[str]:
    complexes_filtered = complex_composition[
        (complex_composition['name_protein'] == element) |
        (complex_composition['gene_name_protein'] == element) |
        (complex_composition['protein_name_protein'] == element) |
        (complex_composition['ensembl_protein'] == element)]

    return complexes_filtered['name_complex'].drop_duplicates().tolist()



import pandas as pd


def call(complexes: pd.DataFrame) -> pd.DataFrame:
    complexes.rename(index=str, columns={'complex_name': 'name'}, inplace=True)

    complexes.rename(index=str, columns={'uniprot_1': 'protein_1', 'uniprot_2': 'protein_2', 'uniprot_3': 'protein_3',
                                         'uniprot_4': 'protein_4'},
                     inplace=True)

    return complexes
